@@ -13,12 +13,19 @@ import numpy as np
 import pandas as pd
 from scipy import optimize
 
+from diff_diff.linalg import solve_ols
 from diff_diff.results import _get_significance_stars
 from diff_diff.utils import (
     compute_confidence_interval,
     compute_p_value,
 )
 
+# Import Rust backend if available (from _backend to avoid circular imports)
+from diff_diff._backend import HAS_RUST_BACKEND, _rust_bootstrap_weights
+
+# Type alias for pre-computed structures
+PrecomputedData = Dict[str, Any]
+
 # =============================================================================
 # Bootstrap Weight Generators
 # =============================================================================
@@ -75,6 +82,98 @@ def _generate_bootstrap_weights(
         )
 
 
+def _generate_bootstrap_weights_batch(
+    n_bootstrap: int,
+    n_units: int,
+    weight_type: str,
+    rng: np.random.Generator,
+) -> np.ndarray:
+    """
+    Generate all bootstrap weights at once (vectorized).
+
+    Parameters
+    ----------
+    n_bootstrap : int
+        Number of bootstrap iterations.
+    n_units : int
+        Number of units (clusters) to generate weights for.
+    weight_type : str
+        Type of weights: "rademacher", "mammen", or "webb".
+    rng : np.random.Generator
+        Random number generator.
+
+    Returns
+    -------
+    np.ndarray
+        Array of bootstrap weights with shape (n_bootstrap, n_units).
+    """
+    # Use Rust backend if available (parallel + fast RNG)
+    if HAS_RUST_BACKEND:
+        # Get seed from the NumPy RNG for reproducibility
+        seed = rng.integers(0, 2**63 - 1)
+        return _rust_bootstrap_weights(n_bootstrap, n_units, weight_type, seed)
+
+    # Fallback to NumPy implementation
+    return _generate_bootstrap_weights_batch_numpy(n_bootstrap, n_units, weight_type, rng)
+
+
+def _generate_bootstrap_weights_batch_numpy(
+    n_bootstrap: int,
+    n_units: int,
+    weight_type: str,
+    rng: np.random.Generator,
+) -> np.ndarray:
+    """
+    NumPy fallback implementation of _generate_bootstrap_weights_batch.
+
+    Generates multiplier bootstrap weights for wild cluster bootstrap.
+    All weight distributions satisfy E[w] = 0, E[w^2] = 1.
+
+    Parameters
+    ----------
+    n_bootstrap : int
+        Number of bootstrap iterations.
+    n_units : int
+        Number of units (clusters) to generate weights for.
+    weight_type : str
+        Type of weights: "rademacher" (+-1), "mammen" (2-point),
+        or "webb" (6-point).
+    rng : np.random.Generator
+        Random number generator for reproducibility.
+
+    Returns
+    -------
+    np.ndarray
+        Array of bootstrap weights with shape (n_bootstrap, n_units).
+    """
+    if weight_type == "rademacher":
+        # Rademacher: +1 or -1 with equal probability
+        return rng.choice([-1.0, 1.0], size=(n_bootstrap, n_units))
+
+    elif weight_type == "mammen":
+        # Mammen's two-point distribution
+        sqrt5 = np.sqrt(5)
+        val1 = -(sqrt5 - 1) / 2
+        val2 = (sqrt5 + 1) / 2
+        p1 = (sqrt5 + 1) / (2 * sqrt5)
+        return rng.choice([val1, val2], size=(n_bootstrap, n_units), p=[p1, 1 - p1])
+
+    elif weight_type == "webb":
+        # Webb's 6-point distribution
+        values = np.array([
+            -np.sqrt(3 / 2), -np.sqrt(2 / 2), -np.sqrt(1 / 2),
+            np.sqrt(1 / 2), np.sqrt(2 / 2), np.sqrt(3 / 2)
+        ])
+        probs = np.array([1, 2, 3, 3, 2, 1]) / 12
+        return rng.choice(values, size=(n_bootstrap, n_units), p=probs)
+
+    else:
+        raise ValueError(
+            f"weight_type must be 'rademacher', 'mammen', or 'webb', "
+            f"got '{weight_type}'"
+        )
+
+
 # =============================================================================
 # Bootstrap Results Container
 # =============================================================================
@@ -226,15 +325,8 @@ def _linear_regression(
     # Add intercept
     X_with_intercept = np.column_stack([np.ones(n), X])
 
-    # OLS: beta = (X'X)^{-1} X'y
-    try:
-        beta = np.linalg.lstsq(X_with_intercept, y, rcond=None)[0]
-    except np.linalg.LinAlgError:
-        # Fallback: use pseudo-inverse
-        beta = np.linalg.pinv(X_with_intercept) @ y
-
-    fitted = X_with_intercept @ beta
-    residuals = y - fitted
+    # Use unified OLS backend (no vcov needed)
+    beta, residuals, _ = solve_ols(X_with_intercept, y, return_vcov=False)
 
     return beta, residuals
 
@@ -553,6 +645,14 @@ class CallawaySantAnna:
         Number of bootstrap iterations for inference.
         If 0, uses analytical standard errors.
         Recommended: 999 or more for reliable inference.
+
+        .. note:: Memory Usage
+            The bootstrap stores all weights in memory as a (n_bootstrap, n_units)
+            float64 array. For large datasets, this can be significant:
+            - 1K bootstrap × 10K units = ~80 MB
+            - 10K bootstrap × 100K units = ~8 GB
+            Consider reducing n_bootstrap if memory is constrained.
+
     bootstrap_weights : str, default="rademacher"
         Type of weights for multiplier bootstrap:
         - "rademacher": +1/-1 with equal probability (standard choice)
@@ -694,6 +794,199 @@ class CallawaySantAnna:
         self.is_fitted_ = False
         self.results_ = None
 
+    def _precompute_structures(
+        self,
+        df: pd.DataFrame,
+        outcome: str,
+        unit: str,
+        time: str,
+        first_treat: str,
+        covariates: Optional[List[str]],
+        time_periods: List[Any],
+        treatment_groups: List[Any],
+    ) -> PrecomputedData:
+        """
+        Pre-compute data structures for efficient ATT(g,t) computation.
+
+        This pivots data to wide format and pre-computes:
+        - Outcome matrix (units x time periods)
+        - Covariate matrix (units x covariates) from base period
+        - Unit cohort membership masks
+        - Control unit masks
+
+        Returns
+        -------
+        PrecomputedData
+            Dictionary with pre-computed structures.
+        """
+        # Get unique units and their cohort assignments
+        unit_info = df.groupby(unit)[first_treat].first()
+        all_units = unit_info.index.values
+        unit_cohorts = unit_info.values
+        n_units = len(all_units)
+
+        # Create unit index mapping for fast lookups
+        unit_to_idx = {u: i for i, u in enumerate(all_units)}
+
+        # Pivot outcome to wide format: rows = units, columns = time periods
+        outcome_wide = df.pivot(index=unit, columns=time, values=outcome)
+        # Reindex to ensure all units are present (handles unbalanced panels)
+        outcome_wide = outcome_wide.reindex(all_units)
+        outcome_matrix = outcome_wide.values  # Shape: (n_units, n_periods)
+        period_to_col = {t: i for i, t in enumerate(outcome_wide.columns)}
+
+        # Pre-compute cohort masks (boolean arrays)
+        cohort_masks = {}
+        for g in treatment_groups:
+            cohort_masks[g] = (unit_cohorts == g)
+
+        # Never-treated mask
+        never_treated_mask = (unit_cohorts == 0) | (unit_cohorts == np.inf)
+
+        # Pre-compute covariate matrices by time period if needed
+        # (covariates are retrieved from the base period of each comparison)
+        covariate_by_period = None
+        if covariates:
+            covariate_by_period = {}
+            for t in time_periods:
+                period_data = df[df[time] == t].set_index(unit)
+                period_cov = period_data.reindex(all_units)[covariates]
+                covariate_by_period[t] = period_cov.values  # Shape: (n_units, n_covariates)
+
+        return {
+            'all_units': all_units,
+            'unit_to_idx': unit_to_idx,
+            'unit_cohorts': unit_cohorts,
+            'outcome_matrix': outcome_matrix,
+            'period_to_col': period_to_col,
+            'cohort_masks': cohort_masks,
+            'never_treated_mask': never_treated_mask,
+            'covariate_by_period': covariate_by_period,
+            'time_periods': time_periods,
+        }
+
+    def _compute_att_gt_fast(
+        self,
+        precomputed: PrecomputedData,
+        g: Any,
+        t: Any,
+        covariates: Optional[List[str]],
+    ) -> Tuple[Optional[float], float, int, int, Optional[Dict[str, Any]]]:
+        """
+        Compute ATT(g,t) using pre-computed data structures (fast version).
+
+        Uses vectorized numpy operations on pre-pivoted outcome matrix
+        instead of repeated pandas filtering.
+        """
+        time_periods = precomputed['time_periods']
+        period_to_col = precomputed['period_to_col']
+        outcome_matrix = precomputed['outcome_matrix']
+        cohort_masks = precomputed['cohort_masks']
+        never_treated_mask = precomputed['never_treated_mask']
+        unit_cohorts = precomputed['unit_cohorts']
+        all_units = precomputed['all_units']
+        covariate_by_period = precomputed['covariate_by_period']
+
+        # Base period for comparison
+        base_period = g - 1 - self.anticipation
+        if base_period not in period_to_col:
+            # Find closest earlier period
+            earlier = [p for p in time_periods if p < g - self.anticipation]
+            if not earlier:
+                return None, 0.0, 0, 0, None
+            base_period = max(earlier)
+
+        # Check if periods exist in the data
+        if base_period not in period_to_col or t not in period_to_col:
+            return None, 0.0, 0, 0, None
+
+        base_col = period_to_col[base_period]
+        post_col = period_to_col[t]
+
+        # Get treated units mask (cohort g)
+        treated_mask = cohort_masks[g]
+
+        # Get control units mask
+        if self.control_group == "never_treated":
+            control_mask = never_treated_mask
+        else:  # not_yet_treated
+            # Not yet treated at time t: never-treated OR first_treat > t
+            control_mask = never_treated_mask | (unit_cohorts > t)
+
+        # Extract outcomes for base and post periods
+        y_base = outcome_matrix[:, base_col]
+        y_post = outcome_matrix[:, post_col]
+
+        # Compute outcome changes (vectorized)
+        outcome_change = y_post - y_base
+
+        # Filter to units with valid data (no NaN in either period)
+        valid_mask = ~(np.isnan(y_base) | np.isnan(y_post))
+
+        # Get treated and control with valid data
+        treated_valid = treated_mask & valid_mask
+        control_valid = control_mask & valid_mask
+
+        n_treated = np.sum(treated_valid)
+        n_control = np.sum(control_valid)
+
+        if n_treated == 0 or n_control == 0:
+            return None, 0.0, 0, 0, None
+
+        # Extract outcome changes for treated and control
+        treated_change = outcome_change[treated_valid]
+        control_change = outcome_change[control_valid]
+
+        # Get unit IDs for influence function
+        treated_units = all_units[treated_valid]
+        control_units = all_units[control_valid]
+
+        # Get covariates if specified (from the base period)
+        X_treated = None
+        X_control = None
+        if covariates and covariate_by_period is not None:
+            cov_matrix = covariate_by_period[base_period]
+            X_treated = cov_matrix[treated_valid]
+            X_control = cov_matrix[control_valid]
+
+            # Check for missing values
+            if np.any(np.isnan(X_treated)) or np.any(np.isnan(X_control)):
+                warnings.warn(
+                    f"Missing values in covariates for group {g}, time {t}. "
+                    "Falling back to unconditional estimation.",
+                    UserWarning,
+                    stacklevel=3,
+                )
+                X_treated = None
+                X_control = None
+
+        # Estimation method
+        if self.estimation_method == "reg":
+            att_gt, se_gt, inf_func = self._outcome_regression(
+                treated_change, control_change, X_treated, X_control
+            )
+        elif self.estimation_method == "ipw":
+            att_gt, se_gt, inf_func = self._ipw_estimation(
+                treated_change, control_change,
+                int(n_treated), int(n_control),
+                X_treated, X_control
+            )
+        else:  # doubly robust
+            att_gt, se_gt, inf_func = self._doubly_robust(
+                treated_change, control_change, X_treated, X_control
+            )
+
+        # Package influence function info with unit IDs for bootstrap
+        n_t = int(n_treated)
+        inf_func_info = {
+            'treated_units': list(treated_units),
+            'control_units': list(control_units),
+            'treated_inf': inf_func[:n_t],
+            'control_inf': inf_func[n_t:],
+        }
+
+        return att_gt, se_gt, int(n_treated), int(n_control), inf_func_info
+
     def fit(
         self,
         data: pd.DataFrame,
@@ -779,6 +1072,12 @@ class CallawaySantAnna:
         if n_control_units == 0:
             raise ValueError("No never-treated units found. Check 'first_treat' column.")
 
+        # Pre-compute data structures for efficient ATT(g,t) computation
+        precomputed = self._precompute_structures(
+            df, outcome, unit, time, first_treat,
+            covariates, time_periods, treatment_groups
+        )
+
         # Compute ATT(g,t) for each group-time combination
         group_time_effects = {}
         influence_func_info = {}  # Store influence functions for bootstrap
@@ -788,9 +1087,8 @@ class CallawaySantAnna:
             valid_periods = [t for t in time_periods if t >= g - self.anticipation]
 
             for t in valid_periods:
-                att_gt, se_gt, n_treat, n_ctrl, inf_info = self._compute_att_gt(
-                    df, outcome, unit, time, first_treat, g, t,
-                    covariates, time_periods
+                att_gt, se_gt, n_treat, n_ctrl, inf_info = self._compute_att_gt_fast(
+                    precomputed, g, t, covariates
                 )
 
                 if att_gt is not None:
@@ -818,7 +1116,9 @@ class CallawaySantAnna:
             )
 
         # Compute overall ATT (simple aggregation)
-        overall_att, overall_se = self._aggregate_simple(group_time_effects, df, unit)
+        overall_att, overall_se = self._aggregate_simple(
+            group_time_effects, influence_func_info, df, unit
+        )
         overall_t = overall_att / overall_se if overall_se > 0 else 0.0
         overall_p = compute_p_value(overall_t)
         overall_ci = compute_confidence_interval(overall_att, overall_se, self.alpha)
@@ -829,12 +1129,13 @@ class CallawaySantAnna:
 
         if aggregate in ["event_study", "all"]:
             event_study_effects = self._aggregate_event_study(
-                group_time_effects, treatment_groups, time_periods, balance_e
+                group_time_effects, influence_func_info,
+                treatment_groups, time_periods, balance_e
             )
 
         if aggregate in ["group", "all"]:
             group_effects = self._aggregate_by_group(
-                group_time_effects, treatment_groups
+                group_time_effects, influence_func_info, treatment_groups
             )
 
         # Run bootstrap inference if requested
@@ -917,135 +1218,6 @@ class CallawaySantAnna:
         self.is_fitted_ = True
         return self.results_
 
-    def _compute_att_gt(
-        self,
-        df: pd.DataFrame,
-        outcome: str,
-        unit: str,
-        time: str,
-        first_treat: str,
-        g: Any,
-        t: Any,
-        covariates: Optional[List[str]],
-        all_periods: List[Any],
-    ) -> Tuple[Optional[float], float, int, int, Optional[Dict[str, Any]]]:
-        """
-        Compute ATT(g,t) for a specific group-time combination.
-
-        Uses 2x2 DiD comparing:
-        - Treated: Units in cohort g
-        - Control: Never-treated units (or not-yet-treated if specified)
-        - Pre-period: g - 1 (or earlier if anticipation > 0)
-        - Post-period: t
-        """
-        # Base period for comparison
-        base_period = g - 1 - self.anticipation
-        if base_period not in all_periods:
-            # Find closest earlier period
-            earlier = [p for p in all_periods if p < g - self.anticipation]
-            if not earlier:
-                return None, 0.0, 0, 0, None
-            base_period = max(earlier)
-
-        # Treated group: units first treated in period g
-        treated_units = df[df[first_treat] == g][unit].unique()
-
-        # Control group
-        if self.control_group == "never_treated":
-            control_mask = df['_never_treated']
-        else:  # not_yet_treated
-            # Not yet treated at time t
-            control_mask = (df['_never_treated']) | (df[first_treat] > t)
-
-        control_units = df[control_mask][unit].unique()
-
-        if len(treated_units) == 0 or len(control_units) == 0:
-            return None, 0.0, 0, 0, None
-
-        # Get data for the two periods
-        df_base = df[df[time] == base_period].set_index(unit)
-        df_post = df[df[time] == t].set_index(unit)
-
-        # Compute outcome changes for treated
-        treated_base = df_base.loc[df_base.index.isin(treated_units), outcome]
-        treated_post = df_post.loc[df_post.index.isin(treated_units), outcome]
-        treated_common = treated_base.index.intersection(treated_post.index)
-
-        if len(treated_common) == 0:
-            return None, 0.0, 0, 0, None
-
-        treated_change = (
-            treated_post.loc[treated_common].values -
-            treated_base.loc[treated_common].values
-        )
-
-        # Compute outcome changes for control
-        control_base = df_base.loc[df_base.index.isin(control_units), outcome]
-        control_post = df_post.loc[df_post.index.isin(control_units), outcome]
-        control_common = control_base.index.intersection(control_post.index)
-
-        if len(control_common) == 0:
-            return None, 0.0, 0, 0, None
-
-        control_change = (
-            control_post.loc[control_common].values -
-            control_base.loc[control_common].values
-        )
-
-        # Get covariates if specified (use base period values for conditioning)
-        X_treated = None
-        X_control = None
-        if covariates:
-            try:
-                X_treated = df_base.loc[treated_common, covariates].values
-                X_control = df_base.loc[control_common, covariates].values
-                # Check for missing values and handle them
-                if np.any(np.isnan(X_treated)) or np.any(np.isnan(X_control)):
-                    warnings.warn(
-                        f"Missing values in covariates for group {g}, time {t}. "
-                        "Falling back to unconditional estimation.",
-                        UserWarning,
-                        stacklevel=3,
-                    )
-                    X_treated = None
-                    X_control = None
-            except KeyError:
-                warnings.warn(
-                    f"Could not extract covariates for group {g}, time {t}. "
-                    "Falling back to unconditional estimation.",
-                    UserWarning,
-                    stacklevel=3,
-                )
-                X_treated = None
-                X_control = None
-
-        # Estimation method
-        if self.estimation_method == "reg":
-            att_gt, se_gt, inf_func = self._outcome_regression(
-                treated_change, control_change, X_treated, X_control
-            )
-        elif self.estimation_method == "ipw":
-            att_gt, se_gt, inf_func = self._ipw_estimation(
-                treated_change, control_change,
-                len(treated_common), len(control_common),
-                X_treated, X_control
-            )
-        else:  # doubly robust
-            att_gt, se_gt, inf_func = self._doubly_robust(
-                treated_change, control_change, X_treated, X_control
-            )
-
-        # Package influence function info with unit IDs for bootstrap
-        n_t = len(treated_common)
-        inf_func_info = {
-            'treated_units': list(treated_common),
-            'control_units': list(control_common),
-            'treated_inf': inf_func[:n_t],
-            'control_inf': inf_func[n_t:],
-        }
-
-        return att_gt, se_gt, len(treated_common), len(control_common), inf_func_info
-
     def _outcome_regression(
         self,
         treated_change: np.ndarray,
@@ -1296,6 +1468,7 @@ class CallawaySantAnna:
     def _aggregate_simple(
         self,
         group_time_effects: Dict,
+        influence_func_info: Dict,
         df: pd.DataFrame,
         unit: str,
     ) -> Tuple[float, float]:
@@ -1303,19 +1476,22 @@ class CallawaySantAnna:
         Compute simple weighted average of ATT(g,t).
 
         Weights by group size (number of treated units).
+
+        Standard errors are computed using influence function aggregation,
+        which properly accounts for covariances across (g,t) pairs due to
+        shared control units. This matches R's `did` package approach.
         """
         effects = []
-        weights = []
-        variances = []
+        weights_list = []
+        gt_pairs = []
 
         for (g, t), data in group_time_effects.items():
             effects.append(data['effect'])
-            weights.append(data['n_treated'])
-            variances.append(data['se'] ** 2)
+            weights_list.append(data['n_treated'])
+            gt_pairs.append((g, t))
 
         effects = np.array(effects)
-        weights = np.array(weights, dtype=float)
-        variances = np.array(variances)
+        weights = np.array(weights_list, dtype=float)
 
         # Normalize weights
         weights = weights / np.sum(weights)
@@ -1323,15 +1499,77 @@ class CallawaySantAnna:
         # Weighted average
         overall_att = np.sum(weights * effects)
 
-        # Standard error (assuming independence across g,t)
-        overall_var = np.sum((weights ** 2) * variances)
-        overall_se = np.sqrt(overall_var)
+        # Compute SE using influence function aggregation
+        overall_se = self._compute_aggregated_se(
+            gt_pairs, weights, influence_func_info
+        )
 
         return overall_att, overall_se
 
+    def _compute_aggregated_se(
+        self,
+        gt_pairs: List[Tuple[Any, Any]],
+        weights: np.ndarray,
+        influence_func_info: Dict,
+    ) -> float:
+        """
+        Compute standard error using influence function aggregation.
+
+        This properly accounts for covariances across (g,t) pairs by
+        aggregating unit-level influence functions:
+
+            ψ_i(overall) = Σ_{(g,t)} w_(g,t) × ψ_i(g,t)
+            Var(overall) = (1/n) Σ_i [ψ_i]²
+
+        This matches R's `did` package analytical SE formula.
+        """
+        if not influence_func_info:
+            # Fallback if no influence functions available
+            return 0.0
+
+        # Build unit index mapping from all (g,t) pairs
+        all_units = set()
+        for (g, t) in gt_pairs:
+            if (g, t) in influence_func_info:
+                info = influence_func_info[(g, t)]
+                all_units.update(info['treated_units'])
+                all_units.update(info['control_units'])
+
+        if not all_units:
+            return 0.0
+
+        all_units = sorted(all_units)
+        n_units = len(all_units)
+        unit_to_idx = {u: i for i, u in enumerate(all_units)}
+
+        # Aggregate influence functions across (g,t) pairs
+        psi_overall = np.zeros(n_units)
+
+        for j, (g, t) in enumerate(gt_pairs):
+            if (g, t) not in influence_func_info:
+                continue
+
+            info = influence_func_info[(g, t)]
+            w = weights[j]
+
+            # Treated unit contributions
+            for i, unit_id in enumerate(info['treated_units']):
+                idx = unit_to_idx[unit_id]
+                psi_overall[idx] += w * info['treated_inf'][i]
+
+            # Control unit contributions
+            for i, unit_id in enumerate(info['control_units']):
+                idx = unit_to_idx[unit_id]
+                psi_overall[idx] += w * info['control_inf'][i]
+
+        # Compute variance: Var(θ̄) = (1/n) Σᵢ ψᵢ²
+        variance = np.sum(psi_overall ** 2)
+        return np.sqrt(variance)
+
     def _aggregate_event_study(
         self,
         group_time_effects: Dict,
+        influence_func_info: Dict,
         groups: List[Any],
         time_periods: List[Any],
         balance_e: Optional[int] = None,
@@ -1340,17 +1578,20 @@ class CallawaySantAnna:
         Aggregate effects by relative time (event study).
 
         Computes average effect at each event time e = t - g.
+
+        Standard errors use influence function aggregation to account for
+        covariances across (g,t) pairs.
         """
-        # Organize effects by relative time
-        effects_by_e: Dict[int, List[Tuple[float, float, int]]] = {}
+        # Organize effects by relative time, keeping track of (g,t) pairs
+        effects_by_e: Dict[int, List[Tuple[Tuple[Any, Any], float, int]]] = {}
 
         for (g, t), data in group_time_effects.items():
             e = t - g  # Relative time
             if e not in effects_by_e:
                 effects_by_e[e] = []
             effects_by_e[e].append((
+                (g, t),  # Keep track of the (g,t) pair
                 data['effect'],
-                data['se'],
                 data['n_treated']
             ))
 
@@ -1363,15 +1604,15 @@ class CallawaySantAnna:
                     groups_at_e.add(g)
 
             # Filter effects to only include balanced groups
-            balanced_effects: Dict[int, List[Tuple[float, float, int]]] = {}
+            balanced_effects: Dict[int, List[Tuple[Tuple[Any, Any], float, int]]] = {}
             for (g, t), data in group_time_effects.items():
                 if g in groups_at_e:
                     e = t - g
                     if e not in balanced_effects:
                         balanced_effects[e] = []
                     balanced_effects[e].append((
+                        (g, t),
                         data['effect'],
-                        data['se'],
                         data['n_treated']
                     ))
             effects_by_e = balanced_effects
@@ -1380,16 +1621,19 @@ class CallawaySantAnna:
         event_study_effects = {}
 
         for e, effect_list in sorted(effects_by_e.items()):
-            effs = np.array([x[0] for x in effect_list])
-            ses = np.array([x[1] for x in effect_list])
+            gt_pairs = [x[0] for x in effect_list]
+            effs = np.array([x[1] for x in effect_list])
             ns = np.array([x[2] for x in effect_list], dtype=float)
 
             # Weight by group size
             weights = ns / np.sum(ns)
 
             agg_effect = np.sum(weights * effs)
-            agg_var = np.sum((weights ** 2) * (ses ** 2))
-            agg_se = np.sqrt(agg_var)
+
+            # Compute SE using influence function aggregation
+            agg_se = self._compute_aggregated_se(
+                gt_pairs, weights, influence_func_info
+            )
 
             t_stat = agg_effect / agg_se if agg_se > 0 else 0.0
             p_val = compute_p_value(t_stat)
@@ -1409,19 +1653,24 @@ class CallawaySantAnna:
     def _aggregate_by_group(
         self,
         group_time_effects: Dict,
+        influence_func_info: Dict,
         groups: List[Any],
     ) -> Dict[Any, Dict[str, Any]]:
         """
         Aggregate effects by treatment cohort.
 
         Computes average effect for each cohort across all post-treatment periods.
+
+        Standard errors use influence function aggregation to account for
+        covariances across time periods within a cohort.
         """
         group_effects = {}
 
         for g in groups:
             # Get all effects for this group (post-treatment only: t >= g)
+            # Keep track of (g, t) pairs for influence function aggregation
             g_effects = [
-                (data['effect'], data['se'], data['n_treated'])
+                ((g, t), data['effect'])
                 for (gg, t), data in group_time_effects.items()
                 if gg == g and t >= g
             ]
@@ -1429,15 +1678,18 @@ class CallawaySantAnna:
             if not g_effects:
                 continue
 
-            effs = np.array([x[0] for x in g_effects])
-            ses = np.array([x[1] for x in g_effects])
+            gt_pairs = [x[0] for x in g_effects]
+            effs = np.array([x[1] for x in g_effects])
 
             # Equal weight across time periods for a group
             weights = np.ones(len(effs)) / len(effs)
 
             agg_effect = np.sum(weights * effs)
-            agg_var = np.sum((weights ** 2) * (ses ** 2))
-            agg_se = np.sqrt(agg_var)
+
+            # Compute SE using influence function aggregation
+            agg_se = self._compute_aggregated_se(
+                gt_pairs, weights, influence_func_info
+            )
 
             t_stat = agg_effect / agg_se if agg_se > 0 else 0.0
             p_val = compute_p_value(t_stat)
@@ -1539,70 +1791,80 @@ class CallawaySantAnna:
                 gt_pairs, group_time_effects, treatment_groups
             )
 
-        # Bootstrap arrays to store results
-        bootstrap_atts_gt = np.zeros((self.n_bootstrap, n_gt))
-        bootstrap_overall = np.zeros(self.n_bootstrap)
+        # Pre-compute unit index arrays for each (g,t) pair (done once, not per iteration)
+        gt_treated_indices = []
+        gt_control_indices = []
+        gt_treated_inf = []
+        gt_control_inf = []
 
+        for j, gt in enumerate(gt_pairs):
+            info = influence_func_info[gt]
+            treated_idx = np.array([unit_to_idx[u] for u in info['treated_units']])
+            control_idx = np.array([unit_to_idx[u] for u in info['control_units']])
+            gt_treated_indices.append(treated_idx)
+            gt_control_indices.append(control_idx)
+            gt_treated_inf.append(np.asarray(info['treated_inf']))
+            gt_control_inf.append(np.asarray(info['control_inf']))
+
+        # Generate ALL bootstrap weights upfront: shape (n_bootstrap, n_units)
+        # This is much faster than generating one at a time
+        all_bootstrap_weights = _generate_bootstrap_weights_batch(
+            self.n_bootstrap, n_units, self.bootstrap_weight_type, rng
+        )
+
+        # Vectorized bootstrap ATT(g,t) computation
+        # Compute all bootstrap ATTs for all (g,t) pairs using matrix operations
+        bootstrap_atts_gt = np.zeros((self.n_bootstrap, n_gt))
+
+        for j in range(n_gt):
+            treated_idx = gt_treated_indices[j]
+            control_idx = gt_control_indices[j]
+            treated_inf = gt_treated_inf[j]
+            control_inf = gt_control_inf[j]
+
+            # Extract weights for this (g,t)'s units across all bootstrap iterations
+            # Shape: (n_bootstrap, n_treated) and (n_bootstrap, n_control)
+            treated_weights = all_bootstrap_weights[:, treated_idx]
+            control_weights = all_bootstrap_weights[:, control_idx]
+
+            # Vectorized perturbation: matrix-vector multiply
+            # Shape: (n_bootstrap,)
+            perturbations = (
+                treated_weights @ treated_inf +
+                control_weights @ control_inf
+            )
+
+            bootstrap_atts_gt[:, j] = original_atts[j] + perturbations
+
+        # Vectorized overall ATT: matrix-vector multiply
+        # Shape: (n_bootstrap,)
+        bootstrap_overall = bootstrap_atts_gt @ overall_weights
+
+        # Vectorized event study aggregation
         if event_study_info is not None:
             rel_periods = sorted(event_study_info.keys())
-            bootstrap_event_study = {e: np.zeros(self.n_bootstrap) for e in rel_periods}
+            bootstrap_event_study = {}
+            for e in rel_periods:
+                agg_info = event_study_info[e]
+                gt_indices = agg_info['gt_indices']
+                weights = agg_info['weights']
+                # Vectorized: select columns and multiply by weights
+                bootstrap_event_study[e] = bootstrap_atts_gt[:, gt_indices] @ weights
         else:
             bootstrap_event_study = None
 
+        # Vectorized group aggregation
         if group_agg_info is not None:
             groups = sorted(group_agg_info.keys())
-            bootstrap_group = {g: np.zeros(self.n_bootstrap) for g in groups}
+            bootstrap_group = {}
+            for g in groups:
+                agg_info = group_agg_info[g]
+                gt_indices = agg_info['gt_indices']
+                weights = agg_info['weights']
+                bootstrap_group[g] = bootstrap_atts_gt[:, gt_indices] @ weights
         else:
             bootstrap_group = None
 
-        # Run bootstrap iterations
-        for b in range(self.n_bootstrap):
-            # Generate unit-level weights
-            unit_weights = _generate_bootstrap_weights(
-                n_units, self.bootstrap_weight_type, rng
-            )
-
-            # Compute bootstrap ATT(g,t) for each group-time pair
-            for j, gt in enumerate(gt_pairs):
-                info = influence_func_info[gt]
-
-                # Get weights for treated and control units
-                treated_indices = [unit_to_idx[u] for u in info['treated_units']]
-                control_indices = [unit_to_idx[u] for u in info['control_units']]
-
-                treated_weights = unit_weights[treated_indices]
-                control_weights = unit_weights[control_indices]
-
-                # Influence function perturbation
-                # Bootstrap ATT* = ATT + sum(weights * influence)
-                perturbation = (
-                    np.sum(treated_weights * info['treated_inf']) +
-                    np.sum(control_weights * info['control_inf'])
-                )
-
-                bootstrap_atts_gt[b, j] = original_atts[j] + perturbation
-
-            # Compute bootstrap overall ATT
-            bootstrap_overall[b] = np.sum(overall_weights * bootstrap_atts_gt[b, :])
-
-            # Compute bootstrap event study effects
-            if bootstrap_event_study is not None and event_study_info is not None:
-                for e, agg_info in event_study_info.items():
-                    gt_indices = agg_info['gt_indices']
-                    weights = agg_info['weights']
-                    bootstrap_event_study[e][b] = np.sum(
-                        weights * bootstrap_atts_gt[b, gt_indices]
-                    )
-
-            # Compute bootstrap group effects
-            if bootstrap_group is not None and group_agg_info is not None:
-                for g, agg_info in group_agg_info.items():
-                    gt_indices = agg_info['gt_indices']
-                    weights = agg_info['weights']
-                    bootstrap_group[g][b] = np.sum(
-                        weights * bootstrap_atts_gt[b, gt_indices]
-                    )
-
         # Compute bootstrap statistics for ATT(g,t)
         gt_ses = {}
         gt_cis = {}
